#include "TclCurveEditor.h"

#include "CurveEditor.h"

#include "Curves.h"

#include "CurveEvaluators.h"

#include <assert.h>

#include <fstream>

#include <map>

static CurveEditor curve_editor;

typedef std::map<std::string, std::string> CurveNameWidgetNameMapping;

CurveNameWidgetNameMapping curve_name_widget_name_mapping;

int __declspec(dllexport) Curveeditor_Init(Tcl_Interp * interpreter)
{
	// I refuse to bow to Tcl's odd capitalization rules :)

	return CurveEditor_Initialization(interpreter);
}

int CurveEditor_Initialization(Tcl_Interp * interpreter)
{
	Tcl_CreateObjCommand(interpreter, "::CurveEditor::resize", CurveEditor_ResizeCommand, NULL, NULL);
	Tcl_CreateObjCommand(interpreter, "::CurveEditor::redraw", CurveEditor_RedrawCommand, NULL, NULL);

	Tcl_CreateObjCommand(interpreter, "::CurveEditor::mouseButtonDown", CurveEditor_MouseButtonDownCommand, NULL, NULL);

	Tcl_CreateObjCommand(interpreter, "::CurveEditor::controlClick", CurveEditor_MouseControlClickCommand, NULL, NULL);

	Tcl_CreateObjCommand(interpreter, "::CurveEditor::altClick", CurveEditor_MouseAltClickCommand, NULL, NULL);
	
	Tcl_CreateObjCommand(interpreter, "::CurveEditor::mouseDrag", CurveEditor_MouseDragCommand, NULL, NULL);
	Tcl_CreateObjCommand(interpreter, "::CurveEditor::mouseButtonUp", CurveEditor_MouseButtonUpCommand, NULL, NULL);

	Tcl_CreateObjCommand(interpreter, "::CurveEditor::shiftMouseButtonDown", CurveEditor_ShiftMouseButtonDownCommand, NULL, NULL);
	Tcl_CreateObjCommand(interpreter, "::CurveEditor::shiftMouseDrag", CurveEditor_ShiftMouseDragCommand, NULL, NULL);
	Tcl_CreateObjCommand(interpreter, "::CurveEditor::shiftMouseButtonUp", CurveEditor_ShiftMouseButtonUpCommand, NULL, NULL);

	Tcl_CreateObjCommand(interpreter, "::CurveEditor::getCurveNames", CurveEditor_GetCurveNamesCommand, NULL, NULL);

	Tcl_CreateObjCommand(interpreter, "::CurveEditor::getCurveTypes", CurveEditor_GetCurveTypesCommand, NULL, NULL);

	Tcl_CreateObjCommand(interpreter, "::CurveEditor::updateCurveCurrentPositions", CurveEditor_UpdateCurveCurrentPositionsCommand, NULL, NULL);

	Tcl_CreateObjCommand(interpreter, "::CurveEditor::setCurrentlySelectedCurveType", CurveEditor_SetCurrentlySelectedCurveTypeCommand, NULL, NULL);
	
	Tcl_CreateObjCommand(interpreter, "::CurveEditor::toggleShowAnimationCurve", CurveEditor_ToggleShowAnimationCurveCommand, NULL, NULL);

	Tcl_CreateObjCommand(interpreter, "::CurveEditor::toggleWrapControlPoints", CurveEditor_ToggleWrapControlPointsCommand, NULL, NULL);
	
	Tcl_CreateObjCommand(interpreter, "::CurveEditor::addCurve", CurveEditor_AddCurveCommand, NULL, NULL);

	Tcl_CreateObjCommand(interpreter, "::CurveEditor::clearCurves", CurveEditor_ClearCurvesCommand, NULL, NULL);

	Tcl_CreateObjCommand(interpreter, "::CurveEditor::loadCurves", CurveEditor_LoadCurvesCommand, NULL, NULL);

	Tcl_CreateObjCommand(interpreter, "::CurveEditor::saveCurves", CurveEditor_SaveCurvesCommand, NULL, NULL);

	Tcl_CreateObjCommand(interpreter, "::CurveEditor::getTime", CurveEditor_GetTimeCommand, NULL, NULL);

	Tcl_CreateObjCommand(interpreter, "::CurveEditor::getCurrentlySelectedCurveName", CurveEditor_GetCurrentlySelectedCurveNameCommand, NULL, NULL);

	Tcl_CreateObjCommand(interpreter, "::CurveEditor::tryPopZoom", CurveEditor_TryPopZoomCommand, NULL, NULL);

	Tcl_CreateObjCommand(interpreter, "::CurveEditor::changeAnimationLength", CurveEditor_ChangeAnimationLengthCommand, NULL, NULL);

	Tcl_CreateObjCommand(interpreter, "::CurveEditor::rescaleTime", CurveEditor_RescaleTimeCommand, NULL, NULL);

	Tcl_CreateObjCommand(interpreter, "::CurveEditor::changeMinimumY", CurveEditor_ChangeMinimumYCommand, NULL, NULL);

	Tcl_CreateObjCommand(interpreter, "::CurveEditor::changeMaximumY", CurveEditor_ChangeMaximumYCommand, NULL, NULL);

	return TCL_OK;
}

int CurveEditor_RedrawCommand(ClientData client_data, Tcl_Interp * interpreter, int object_count, Tcl_Obj * const objects[])
{
	int result;

	Tcl_Obj * current_time_object = Tcl_NewStringObj("g_currtime", -1);

	double current_time;

	if (Tcl_GetDoubleFromObj(interpreter, Tcl_ObjGetVar2(interpreter, current_time_object, NULL, TCL_GLOBAL_ONLY ), &current_time) != TCL_OK )
	{
		result = TCL_ERROR;
	}
	else
	{
		result = TCL_OK;
	}
		
	curve_editor.drawCurveEditor(current_time);

	return TCL_OK;
}

int CurveEditor_ResizeCommand(ClientData client_data, Tcl_Interp * interpreter, int object_count, Tcl_Obj * const objects[])
{
	int result;
	
	if (object_count != 3)
    {
		Tcl_SetResult( interpreter, "wrong number of arguments: should be \"width height\"", TCL_STATIC);
		
		result = TCL_ERROR;
    }
	else
	{
		int window_width;
		int window_height;
		
		if (Tcl_GetIntFromObj(interpreter, objects[1], &window_width) != TCL_OK)
		{
			result = TCL_ERROR;
		}
		else
		{
		    if (Tcl_GetIntFromObj(interpreter, objects[2], &window_height) != TCL_OK)
			{
				result = TCL_ERROR;
			}
			else
			{
				curve_editor.specifyWindowSize(window_width, window_height);

				result = TCL_OK;
			}
		}
	}

    return result;
}

int CurveEditor_MouseButtonDownCommand(ClientData client_data, Tcl_Interp * interpreter, int object_count, Tcl_Obj * const objects[])
{
	int result;
	
	if (object_count != 3)
    {
		Tcl_SetResult( interpreter, "wrong number of arguments: should be \"x y\"", TCL_STATIC);
		
		result = TCL_ERROR;
    }
	else
	{
		int x_location;
		int y_location;
		
		if (Tcl_GetIntFromObj(interpreter, objects[1], &x_location) != TCL_OK)
		{
			result = TCL_ERROR;
		}
		else
		{
		    if (Tcl_GetIntFromObj(interpreter, objects[2], &y_location) != TCL_OK)
			{
				result = TCL_ERROR;
			}
			else
			{
				curve_editor.mouseDown(x_location, y_location);

				CurveEditor_UpdateTclVariables(interpreter);

				result = TCL_OK;
			}
		}
	}

    return result;
}

int CurveEditor_MouseAltClickCommand(ClientData client_data, Tcl_Interp * interpreter, int object_count, Tcl_Obj * const objects[])
{
	int result;
	
	if (object_count != 3)
    {
		Tcl_SetResult( interpreter, "wrong number of arguments: should be \"x y\"", TCL_STATIC);
		
		result = TCL_ERROR;
    }
	else
	{
		int x_location;
		int y_location;
		
		if (Tcl_GetIntFromObj(interpreter, objects[1], &x_location) != TCL_OK)
		{
			result = TCL_ERROR;
		}
		else
		{
		    if (Tcl_GetIntFromObj(interpreter, objects[2], &y_location) != TCL_OK)
			{
				result = TCL_ERROR;
			}
			else
			{
				curve_editor.altClick(x_location, y_location);

				result = TCL_OK;
			}
		}
	}

    return result;
}

int CurveEditor_MouseControlClickCommand(ClientData client_data, Tcl_Interp * interpreter, int object_count, Tcl_Obj * const objects[])
{
	int result;
	
	if (object_count != 3)
    {
		Tcl_SetResult( interpreter, "wrong number of arguments: should be \"x y\"", TCL_STATIC);
		
		result = TCL_ERROR;
    }
	else
	{
		int x_location;
		int y_location;
		
		if (Tcl_GetIntFromObj(interpreter, objects[1], &x_location) != TCL_OK)
		{
			result = TCL_ERROR;
		}
		else
		{
		    if (Tcl_GetIntFromObj(interpreter, objects[2], &y_location) != TCL_OK)
			{
				result = TCL_ERROR;
			}
			else
			{
				curve_editor.controlClick(x_location, y_location);

				result = TCL_OK;
			}
		}
	}

    return result;
}

int CurveEditor_MouseDragCommand(ClientData client_data, Tcl_Interp * interpreter, int object_count, Tcl_Obj * const objects[])
{
	int result;
	
	if (object_count != 3)
    {
		Tcl_SetResult( interpreter, "wrong number of arguments: should be \"x y\"", TCL_STATIC);
		
		result = TCL_ERROR;
    }
	else
	{
		int x_location;
		int y_location;
		
		if (Tcl_GetIntFromObj(interpreter, objects[1], &x_location) != TCL_OK)
		{
			result = TCL_ERROR;
		}
		else
		{
		    if (Tcl_GetIntFromObj(interpreter, objects[2], &y_location) != TCL_OK)
			{
				result = TCL_ERROR;
			}
			else
			{
				curve_editor.mouseMove(x_location, y_location);

				result = TCL_OK;
			}
		}
	}

    return result;
}


int CurveEditor_MouseButtonUpCommand(ClientData client_data, Tcl_Interp * interpreter, int object_count, Tcl_Obj * const objects[])
{
	int result;
	
	if (object_count != 3)
    {
		Tcl_SetResult( interpreter, "wrong number of arguments: should be \"x y\"", TCL_STATIC);
		
		result = TCL_ERROR;
    }
	else
	{
		int x_location;
		int y_location;
		
		if (Tcl_GetIntFromObj(interpreter, objects[1], &x_location) != TCL_OK)
		{
			result = TCL_ERROR;
		}
		else
		{
		    if (Tcl_GetIntFromObj(interpreter, objects[2], &y_location) != TCL_OK)
			{
				result = TCL_ERROR;
			}
			else
			{
				curve_editor.mouseUp(x_location, y_location);

				CurveEditor_UpdateTclVariables(interpreter);

				result = TCL_OK;
			}
		}
	}

    return result;
}

int CurveEditor_ShiftMouseButtonDownCommand(ClientData client_data, Tcl_Interp * interpreter, int object_count, Tcl_Obj * const objects[])
{
	int result;
	
	if (object_count != 3)
    {
		Tcl_SetResult( interpreter, "wrong number of arguments: should be \"x y\"", TCL_STATIC);
		
		result = TCL_ERROR;
    }
	else
	{
		int x_location;
		int y_location;
		
		if (Tcl_GetIntFromObj(interpreter, objects[1], &x_location) != TCL_OK)
		{
			result = TCL_ERROR;
		}
		else
		{
		    if (Tcl_GetIntFromObj(interpreter, objects[2], &y_location) != TCL_OK)
			{
				result = TCL_ERROR;
			}
			else
			{
				curve_editor.shiftMouseDown(x_location, y_location);

				result = TCL_OK;
			}
		}
	}

    return result;
}

int CurveEditor_ShiftMouseDragCommand(ClientData client_data, Tcl_Interp * interpreter, int object_count, Tcl_Obj * const objects[])
{
	int result;
	
	if (object_count != 3)
    {
		Tcl_SetResult( interpreter, "wrong number of arguments: should be \"x y\"", TCL_STATIC);
		
		result = TCL_ERROR;
    }
	else
	{
		int x_location;
		int y_location;
		
		if (Tcl_GetIntFromObj(interpreter, objects[1], &x_location) != TCL_OK)
		{
			result = TCL_ERROR;
		}
		else
		{
		    if (Tcl_GetIntFromObj(interpreter, objects[2], &y_location) != TCL_OK)
			{
				result = TCL_ERROR;
			}
			else
			{
				curve_editor.shiftMouseDrag(x_location, y_location);

				result = TCL_OK;
			}
		}
	}

    return result;
}

int CurveEditor_ShiftMouseButtonUpCommand(ClientData client_data, Tcl_Interp * interpreter, int object_count, Tcl_Obj * const objects[])
{
	int result;
	
	if (object_count != 3)
    {
		Tcl_SetResult( interpreter, "wrong number of arguments: should be \"x y\"", TCL_STATIC);
		
		result = TCL_ERROR;
    }
	else
	{
		int x_location;
		int y_location;
		
		if (Tcl_GetIntFromObj(interpreter, objects[1], &x_location) != TCL_OK)
		{
			result = TCL_ERROR;
		}
		else
		{
		    if (Tcl_GetIntFromObj(interpreter, objects[2], &y_location) != TCL_OK)
			{
				result = TCL_ERROR;
			}
			else
			{
				curve_editor.shiftMouseUp(x_location, y_location);

				CurveEditor_UpdateTclVariables(interpreter);

				result = TCL_OK;
			}
		}
	}

    return result;
}

int CurveEditor_GetCurveNamesCommand(ClientData client_data, Tcl_Interp * interpreter, int object_count, Tcl_Obj * const objects[])
{
	int result;

	if (object_count != 1)
	{
		Tcl_SetResult(interpreter, "wrong # args", TCL_STATIC);

		result = TCL_ERROR;
	}
	else
	{
		Curves * curves = Curves::getInstance();

		for (Curves::const_iterator curve_iterator = curves->begin(); curve_iterator != curves->end(); ++curve_iterator)
		{
			std::string curve_name = curve_iterator->getName();

			Tcl_ListObjAppendElement(interpreter, Tcl_GetObjResult(interpreter), Tcl_NewStringObj(curve_name.c_str(), -1));
		}

	    result = TCL_OK;
	}

	return result;
}

int CurveEditor_GetCurveTypesCommand(ClientData client_data, Tcl_Interp * interpreter, int object_count, Tcl_Obj * const objects[])
{
	int result;

	if (object_count != 1)
	{
		Tcl_SetResult(interpreter, "wrong # args", TCL_STATIC);

		result = TCL_ERROR;
	}
	else
	{
		CurveEvaluators * curve_evaluators = CurveEvaluators::getInstance();

		for (CurveEvaluators::const_iterator curve_evaluators_iterator = curve_evaluators->begin(); curve_evaluators_iterator != curve_evaluators->end(); ++curve_evaluators_iterator)
		{
			std::string curve_type = curve_evaluators_iterator->first;

			Tcl_ListObjAppendElement(interpreter, Tcl_GetObjResult(interpreter), Tcl_NewStringObj(curve_type.c_str(), -1));
		}

	    result = TCL_OK;
	}

	return result;
}

int CurveEditor_UpdateCurveCurrentPositionsCommand(ClientData client_data, Tcl_Interp * interpreter, int object_count, Tcl_Obj * const objects[])
{
	int result = TCL_OK;

    if (object_count != 3)
    {
		Tcl_SetResult(interpreter, "wrong # args: should be \"time arrayname\"", TCL_STATIC);

		result = TCL_ERROR;
    }
	else
	{
		double current_time;	

		if (Tcl_GetDoubleFromObj(interpreter, objects[1], &current_time) != TCL_OK)
		{
			result = TCL_ERROR;
		}
		else
		{
			Curves * curves = Curves::getInstance();

			for (Curves::const_iterator curve_iterator = curves->begin(); curve_iterator != curves->end(); ++curve_iterator)
			{
				double current_value = curve_iterator->evaluateCurveAt(current_time);

				CurveNameWidgetNameMapping::const_iterator find_results = curve_name_widget_name_mapping.find(curve_iterator->getName());

				assert(find_results != curve_name_widget_name_mapping.end());

				std::string widget_name = find_results->second;

				Tcl_Obj * set_var_result = Tcl_ObjSetVar2(interpreter, objects[2], Tcl_NewStringObj(widget_name.c_str(), -1), Tcl_NewDoubleObj(current_value), 0 );

				assert(set_var_result != NULL);
			}
		}

	}

	return result;
}

// will need these two

// set current curve type (from menu)

// needs to go to editor to get currently selected curve

int CurveEditor_SetCurrentlySelectedCurveTypeCommand(ClientData client_data, Tcl_Interp * interpreter, int object_count, Tcl_Obj * const objects[])
{
	int result;

    if (object_count != 2)
    {
		Tcl_SetResult(interpreter, "wrong # args", TCL_STATIC);

		result = TCL_ERROR;
    }
	else
	{
		char * curve_type = Tcl_GetString(objects[1]);

		Curve * currently_selected_curve = curve_editor.getCurrentlySelectedCurve();
		
		if (currently_selected_curve != NULL)
		{
			currently_selected_curve->setCurveType(curve_type);
		}

		CurveEditor_UpdateTclVariables(interpreter);

		result = TCL_OK;
	}

	return result;
}

// toggle the usage of an animation variable

// tcl only needs to get editor to update screen, but that shouldn't involve it

int CurveEditor_ToggleShowAnimationCurveCommand(ClientData client_data, Tcl_Interp * interpreter, int object_count, Tcl_Obj * const objects[])
{
	int result;

    if (object_count != 2)
    {
		Tcl_SetResult(interpreter, "wrong # args", TCL_STATIC);

		result = TCL_ERROR;
    }
	else
	{
		char * curve_name;

		curve_name = Tcl_GetString(objects[1]);

		curve_editor.toggleShowAnimationCurve(curve_name);

		CurveEditor_UpdateTclVariables(interpreter);
		
		result = TCL_OK;
	}

	return result;
}


int CurveEditor_ToggleWrapControlPointsCommand(ClientData client_data, Tcl_Interp * interpreter, int object_count, Tcl_Obj * const objects[])
{
	int result;

    if (object_count != 1)
    {
		Tcl_SetResult(interpreter, "wrong # args", TCL_STATIC);

		result = TCL_ERROR;
    }
	else
	{
		curve_editor.toggleWrapControlPoints();

		CurveEditor_UpdateTclVariables(interpreter);
		
		result = TCL_OK;
	}

	return result;
}


int CurveEditor_AddCurveCommand(ClientData client_data, Tcl_Interp * interpreter, int object_count, Tcl_Obj * const objects[])
{
	int result;

    if (object_count != 6)
    {
		Tcl_SetResult(interpreter, "wrong # args", TCL_STATIC);

		result = TCL_ERROR;
    }
	else
	{
		char * curve_name = Tcl_GetString(objects[1]);

		char * widget_name = Tcl_GetString(objects[2]);

		double minimum_value;

		if (Tcl_GetDoubleFromObj(interpreter, objects[3], &minimum_value) != TCL_OK)
		{
			result = TCL_ERROR;
		}
		else
		{
			double maximum_value;

			if (Tcl_GetDoubleFromObj(interpreter, objects[4], &maximum_value) != TCL_OK)
			{
				result = TCL_ERROR;
			}
			else
			{

				double start_value;
				
				if (Tcl_GetDoubleFromObj(interpreter, objects[5], &start_value) != TCL_OK)
				{
					result = TCL_ERROR;
				}
				else
				{
					Curves::getInstance()->addCurve(curve_name, minimum_value, maximum_value, start_value);

					curve_name_widget_name_mapping.insert(CurveNameWidgetNameMapping::value_type(curve_name, widget_name));

					result = TCL_OK;
				}
			}
		}
	}

	return result;
}

int CurveEditor_ClearCurvesCommand(ClientData client_data, Tcl_Interp * interpreter, int object_count, Tcl_Obj * const objects[])
{
	int result = TCL_OK;
	
	curve_editor.resetCurveEditor();

	return result;
}

int CurveEditor_LoadCurvesCommand(ClientData client_data, Tcl_Interp * interpreter, int object_count, Tcl_Obj * const objects[])
{
	int result = TCL_OK;

    if (object_count != 2)
    {
		Tcl_SetResult(interpreter, "wrong # args", TCL_STATIC);

		result = TCL_ERROR;
    }
	else
	{
		char * file_name;

		file_name = Tcl_GetString(objects[1]);

		std::ifstream load_file(file_name);

		Curves::getInstance()->fromStream(load_file);

		load_file.close();

		result = TCL_OK;
	}
	
	return result;
}

int CurveEditor_SaveCurvesCommand(ClientData client_data, Tcl_Interp * interpreter, int object_count, Tcl_Obj * const objects[])
{
	int result = TCL_OK;

    if (object_count != 2)
    {
		Tcl_SetResult(interpreter, "wrong # args", TCL_STATIC);

		result = TCL_ERROR;
    }
	else
	{
		char * file_name;

		file_name = Tcl_GetString(objects[1]);

		std::ofstream save_file(file_name);

		Curves::getInstance()->toStream(save_file);

		save_file.close();

		result = TCL_OK;
	}
	
	return result;
}

int CurveEditor_GetTimeCommand(ClientData client_data, Tcl_Interp * interpreter, int object_count, Tcl_Obj * const objects[])
{
	int result = TCL_OK;

    if (object_count != 2)
    {
		Tcl_SetResult(interpreter, "wrong # args", TCL_STATIC);

		result = TCL_ERROR;
    }
	else
	{
		int x;
		
		if (Tcl_GetIntFromObj(interpreter, objects[1], &x) != TCL_OK)
		{
			result = TCL_ERROR;
		}
		else
		{
			Tcl_SetObjResult(interpreter, Tcl_NewDoubleObj(curve_editor.getTime(x)));

			result = TCL_OK;
		}
	}

	return result;
}

int CurveEditor_GetCurrentlySelectedCurveNameCommand(ClientData client_data, Tcl_Interp * interpreter, int object_count, Tcl_Obj * const objects[])
{
	int result = TCL_OK;

    if (object_count != 1)
    {
		Tcl_SetResult(interpreter, "wrong # args", TCL_STATIC);

		result = TCL_ERROR;
    }
	else
	{
		Curve * currently_selected_curve = curve_editor.getCurrentlySelectedCurve();
		
		std::string currently_selected_curve_name;

		if (currently_selected_curve != NULL)
		{
			currently_selected_curve_name = currently_selected_curve->getName();
		}

		Tcl_SetObjResult(interpreter, Tcl_NewStringObj(currently_selected_curve_name.c_str(), -1));

		result = TCL_OK;
	}

	return result;
}

int CurveEditor_TryPopZoomCommand(ClientData client_data, Tcl_Interp * interpreter, int object_count, Tcl_Obj * const objects[])
{
	int result = TCL_OK;

    if (object_count != 1)
    {
		Tcl_SetResult(interpreter, "wrong # args", TCL_STATIC);

		result = TCL_ERROR;
    }
	else
	{
		curve_editor.tryPopZoom();

		CurveEditor_UpdateTclVariables(interpreter);

		result = TCL_OK;
	}

	return result;
}

int CurveEditor_RescaleTimeCommand(ClientData client_data, Tcl_Interp * interpreter, int object_count, Tcl_Obj * const objects[])
{
	int result = TCL_OK;

    if (object_count != 2)
    {
		Tcl_SetResult(interpreter, "wrong # args", TCL_STATIC);

		result = TCL_ERROR;
    }
	else
	{
		double scaling_factor;	

		if (Tcl_GetDoubleFromObj(interpreter, objects[1], &scaling_factor) != TCL_OK)
		{
			result = TCL_ERROR;
		}
		else
		{
			curve_editor.rescaleTime(scaling_factor);
		}

	}

	return result;
}

int CurveEditor_ChangeAnimationLengthCommand(ClientData client_data, Tcl_Interp * interpreter, int object_count, Tcl_Obj * const objects[])
{
	int result = TCL_OK;

    if (object_count != 2)
    {
		Tcl_SetResult(interpreter, "wrong # args", TCL_STATIC);

		result = TCL_ERROR;
    }
	else
	{
		double animation_length;
		
		if (Tcl_GetDoubleFromObj(interpreter, objects[1], &animation_length) != TCL_OK)
		{
			result = TCL_ERROR;
		}
		else
		{
			curve_editor.changeAnimationLength(animation_length);

			CurveEditor_UpdateTclVariables(interpreter);

			result = TCL_OK;
		}
	}

	return result;
}

int CurveEditor_ChangeMinimumYCommand(ClientData client_data, Tcl_Interp * interpreter, int object_count, Tcl_Obj * const objects[])
{
	int result = TCL_OK;

    if (object_count != 2)
    {
		Tcl_SetResult(interpreter, "wrong # args", TCL_STATIC);

		result = TCL_ERROR;
    }
	else
	{
		double new_minimum_y;
		
		if (Tcl_GetDoubleFromObj(interpreter, objects[1], &new_minimum_y) != TCL_OK)
		{
			result = TCL_ERROR;
		}
		else
		{
			curve_editor.changeMinimumY(new_minimum_y);

			CurveEditor_UpdateTclVariables(interpreter);

			result = TCL_OK;
		}
	}

	return result;
}

int CurveEditor_ChangeMaximumYCommand(ClientData client_data, Tcl_Interp * interpreter, int object_count, Tcl_Obj * const objects[])
{
	int result = TCL_OK;

    if (object_count != 2)
    {
		Tcl_SetResult(interpreter, "wrong # args", TCL_STATIC);

		result = TCL_ERROR;
    }
	else
	{
		double new_maximum_y;
		
		if (Tcl_GetDoubleFromObj(interpreter, objects[1], &new_maximum_y) != TCL_OK)
		{
			result = TCL_ERROR;
		}
		else
		{
			curve_editor.changeMaximumY(new_maximum_y);

			CurveEditor_UpdateTclVariables(interpreter);

			result = TCL_OK;
		}
	}

	return result;
}

void CurveEditor_UpdateTclVariables(Tcl_Interp * interpreter)
{
	double x_start;
	double x_division_interval;
	double x_end;

	double minimum_time = curve_editor.getMinimumTime();

	curve_editor.findNiceDivisions(minimum_time, curve_editor.getMaximumTime(), &x_start, &x_division_interval, &x_end);

	double shift_x = x_start - minimum_time;

	double y_start;
	double y_division_interval;
	double y_end;

	double minimum_value = curve_editor.getMinimumValue();

	curve_editor.findNiceDivisions(minimum_value, curve_editor.getMaximumValue(), &y_start, &y_division_interval, &y_end);

	double shift_y = y_start - minimum_value;

	double new_y_minimum = curve_editor.getMinimumValue();
	double new_y_maximum = curve_editor.getMaximumValue();


	Curve * currently_selected_curve = curve_editor.getCurrentlySelectedCurve();

	unsigned int number_of_control_points = 0;
	unsigned int number_of_curve_segments = 0;

	if (currently_selected_curve != NULL)
	{
		number_of_control_points = currently_selected_curve->getNumberOfControlPoints();

		number_of_curve_segments = currently_selected_curve->getNumberOfEvaluatedCurveSegments();
	}

	Tcl_Obj * return_value;
	
	return_value = Tcl_ObjSetVar2(interpreter, Tcl_NewStringObj("::minimum_time",				-1), NULL, Tcl_NewDoubleObj(x_start				), TCL_GLOBAL_ONLY);
	return_value = Tcl_ObjSetVar2(interpreter, Tcl_NewStringObj("::time_interval",				-1), NULL, Tcl_NewDoubleObj(x_division_interval	), TCL_GLOBAL_ONLY);
	return_value = Tcl_ObjSetVar2(interpreter, Tcl_NewStringObj("::maximum_time",				-1), NULL, Tcl_NewDoubleObj(x_end				), TCL_GLOBAL_ONLY);
	return_value = Tcl_ObjSetVar2(interpreter, Tcl_NewStringObj("::minimum_value",				-1), NULL, Tcl_NewDoubleObj(y_start				), TCL_GLOBAL_ONLY);
	return_value = Tcl_ObjSetVar2(interpreter, Tcl_NewStringObj("::value_interval",				-1), NULL, Tcl_NewDoubleObj(y_division_interval	), TCL_GLOBAL_ONLY);
	return_value = Tcl_ObjSetVar2(interpreter, Tcl_NewStringObj("::maximum_value",				-1), NULL, Tcl_NewDoubleObj(y_end				), TCL_GLOBAL_ONLY);

	return_value = Tcl_ObjSetVar2(interpreter, Tcl_NewStringObj("::shift_x",					-1), NULL, Tcl_NewDoubleObj(shift_x				), TCL_GLOBAL_ONLY);
	return_value = Tcl_ObjSetVar2(interpreter, Tcl_NewStringObj("::shift_y",					-1), NULL, Tcl_NewDoubleObj(shift_y				), TCL_GLOBAL_ONLY);

	return_value = Tcl_ObjSetVar2(interpreter, Tcl_NewStringObj("::new_y_minimum",				-1), NULL, Tcl_NewDoubleObj(new_y_minimum		), TCL_GLOBAL_ONLY);
	return_value = Tcl_ObjSetVar2(interpreter, Tcl_NewStringObj("::new_y_maximum",				-1), NULL, Tcl_NewDoubleObj(new_y_maximum		), TCL_GLOBAL_ONLY);

	return_value = Tcl_ObjSetVar2(interpreter, Tcl_NewStringObj("::number_of_control_points",	-1), NULL, Tcl_NewIntObj(number_of_control_points), TCL_GLOBAL_ONLY);
	return_value = Tcl_ObjSetVar2(interpreter, Tcl_NewStringObj("::number_of_curve_segments",	-1), NULL, Tcl_NewIntObj(number_of_curve_segments), TCL_GLOBAL_ONLY);
}
